import abc
from typing import Generic, TypeVar

from notion_oauth_handler.core.dto import AuthRedirectInfo, TokenResponseInfo


_STATE_TV = TypeVar('_STATE_TV')


class NotionOAuthConsumer(abc.ABC, Generic[_STATE_TV]):
    """
    Class for application-specific implementation of the oauth handling process.

    Requires the implementation of two methods in subclasses:
    - consume_redirect_info
    - consume_token_info

    and, optionally:
    - consume_redirect_error
    """

    async def consume_redirect_error(self, error: str) -> None:
        pass

    @abc.abstractmethod
    async def consume_redirect_info(self, redirect_info: AuthRedirectInfo) -> _STATE_TV:
        raise NotImplementedError

    @abc.abstractmethod
    async def consume_token_info(self, token_info: TokenResponseInfo, state_info: _STATE_TV) -> None:
        raise NotImplementedError


class DefaultNotionOAuthConsumer(NotionOAuthConsumer[AuthRedirectInfo]):
    """
    Default half-implementation; uses `AuthRedirectInfo` as state_info.
    Does not require re-implementation of `consume_redirect_info`
    in subclasses.
    """

    async def consume_redirect_info(self, redirect_info: AuthRedirectInfo) -> AuthRedirectInfo:
        return redirect_info

    @abc.abstractmethod
    async def consume_token_info(self, token_info: TokenResponseInfo, state_info: AuthRedirectInfo) -> None:
        raise NotImplementedError


class DummyNotionOAuthConsumer(DefaultNotionOAuthConsumer):
    async def consume_token_info(self, token_info: TokenResponseInfo, state_info: AuthRedirectInfo) -> None:
        print(f'Consumed token info: {token_info}')



import logging
from http import HTTPStatus

import aiohttp.web

from notion_oauth_handler.core import exc
from notion_oauth_handler.core.dto import AuthRedirectInfo
from notion_oauth_handler.core.oauth_handler import NotionOAuthHandler
from notion_oauth_handler.server.middleware import (
    OAUTH_HANDLER_REQUEST_KEY, RESPONSE_FACTORY_REQUEST_KEY,
)
from notion_oauth_handler.server.response import NotionOAuthResponseFactory


_LOGGER = logging.getLogger(__name__)


class NotionOAuthRedirectView(aiohttp.web.View):
    def get_oauth_handler(self) -> NotionOAuthHandler:
        handler = self.request[OAUTH_HANDLER_REQUEST_KEY]
        assert isinstance(handler, NotionOAuthHandler)
        return handler

    def get_response_factory(self) -> NotionOAuthResponseFactory:
        response_factory = self.request[RESPONSE_FACTORY_REQUEST_KEY]
        assert isinstance(response_factory, NotionOAuthResponseFactory)
        return response_factory

    async def get(self) -> aiohttp.web.Response:
        _LOGGER.info('Accepted redirect request')

        handler = self.get_oauth_handler()
        response_factory = self.get_response_factory()

        error = self.request.query.get('error', '')
        if error:
            try:
                await handler.handle_error(error=error)
            except exc.NotionAccessDenied:
                return await response_factory.make_error_response(error=error)

        redirect_info = AuthRedirectInfo(
            redirect_uri=str(self.request.url),
            state=self.request.query.get('state', ''),
            code=self.request.query.get('code', ''),
        )
        try:
            token_info = await handler.handle_auth(redirect_info=redirect_info)
        except exc.TokenRequestFailed:
            return aiohttp.web.Response(
                status=HTTPStatus.BAD_REQUEST,
                text='Token request failed',
            )

        return await response_factory.make_auth_response(token_info=token_info)



import argparse
import asyncio
import json
import logging
import os
from importlib import metadata
from typing import Any, Type

from aiohttp import web

import notion_oauth_handler as package
from notion_oauth_handler.core.dto import TokenResponseInfo
from notion_oauth_handler.core.consumer import NotionOAuthConsumer
from notion_oauth_handler.server.response import NotionOAuthResponseFactory
from notion_oauth_handler.server.app import make_app
from notion_oauth_handler.mock.app import make_mock_app


def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser('Notion OAuth Handler Server')
    subparsers = parser.add_subparsers(title='command', dest='command')

    response_factory_arg_parser = argparse.ArgumentParser(add_help=False)
    response_factory_arg_parser.add_argument(
        '--response-factory', default='dummy',
        help='Response factory class entrypoint name',
    )

    host_port_arg_parser = argparse.ArgumentParser(add_help=False)
    host_port_arg_parser.add_argument('--host', default='0.0.0.0', help='Server host')
    host_port_arg_parser.add_argument('--port', default='8000', type=int, help='Server port')

    serve_cmd_parser = subparsers.add_parser(
        'serve', help='Run HTTP server',
        parents=[response_factory_arg_parser, host_port_arg_parser]
    )
    serve_cmd_parser.add_argument(
        '--client-id-key', default='NOTION_CLIENT_ID',
        help='Name of env var for Notion client ID',
    )
    serve_cmd_parser.add_argument(
        '--client-secret-key', default='NOTION_CLIENT_SECRET',
        help='Name of env var for Notion client sercret',
    )
    serve_cmd_parser.add_argument(
        '--consumer', default='dummy',
        help='Consumer class entrypoint name',
    )
    serve_cmd_parser.add_argument(
        '--base-path', default='', help='Base path for all endpoints')
    serve_cmd_parser.add_argument(
        '--redirect-path', default='/auth_redirect', help='Path of redirect endpoint')

    subparsers.add_parser(
        'mock', help='Run Notion mock server',
        parents=[host_port_arg_parser]
    )

    consumer_cmd_parser = subparsers.add_parser('consumer', help='Consumer information')
    consumer_cmd_subparsers = consumer_cmd_parser.add_subparsers(
        title='consumer_command', dest='consumer_command')
    consumer_cmd_subparsers.add_parser('list', help='List available consumers')

    response_factory_cmd_parser = subparsers.add_parser('response_factory', help='Response factory information')
    response_factory_cmd_subparsers = response_factory_cmd_parser.add_subparsers(
        title='response_factory_command', dest='response_factory_command')
    response_factory_cmd_subparsers.add_parser('list', help='List available response factories')

    response_cmd_parser = subparsers.add_parser(
        'response', help='Generate server response', parents=[response_factory_arg_parser]
    )
    response_cmd_subparsers = response_cmd_parser.add_subparsers(
        title='response_command', dest='response_command')

    response_auth_cmd_parser = response_cmd_subparsers.add_parser(
        'auth', help='Generate successful auth response')
    response_auth_cmd_parser.add_argument('--access-token', default='<access_token>', help='access_token value')
    response_auth_cmd_parser.add_argument('--workspace-id', default='<workspace_id>', help='workspace_id value')
    response_auth_cmd_parser.add_argument('--workspace-name', default='My Workspace', help='workspace_name value')
    response_auth_cmd_parser.add_argument('--workspace-icon', default='workspace.icon', help='workspace_icon value')
    response_auth_cmd_parser.add_argument('--bot-id', default='<bot_id>', help='bot_id value')
    response_auth_cmd_parser.add_argument('--owner', default='{}', help='owner value')

    response_error_cmd_parser = response_cmd_subparsers.add_parser(
        'error', help='Generate auth error response')
    response_error_cmd_parser.add_argument('--error', default='Error text', help='error value')

    return parser


class NotionOAuthTool:
    @classmethod
    @property
    def _consumer_entrypoint_key(cls) -> str:
        return f'{package.__name__}.consumer'

    @classmethod
    @property
    def _response_factory_entrypoint_key(cls) -> str:
        return f'{package.__name__}.response_factory'

    @classmethod
    def consumer_list(cls) -> None:
        entrypoints = metadata.entry_points()[cls._consumer_entrypoint_key]
        for ep in entrypoints:
            print(ep.name)

    @classmethod
    def response_factory_list(cls) -> None:
        entrypoints = metadata.entry_points()[cls._response_factory_entrypoint_key]
        for ep in entrypoints:
            print(ep.name)

    @classmethod
    def _get_consumer(cls, consumer_name: str) -> NotionOAuthConsumer:
        entrypoints = metadata.entry_points()[cls._consumer_entrypoint_key]
        consumer_cls: Type[NotionOAuthConsumer] = next(
            iter(ep for ep in entrypoints if ep.name == consumer_name)
        ).load()
        consumer = consumer_cls()
        return consumer

    @classmethod
    def _get_response_factory(cls, response_factory_name: str) -> NotionOAuthResponseFactory:
        entrypoints = metadata.entry_points()[cls._response_factory_entrypoint_key]
        response_factory_cls: Type[NotionOAuthResponseFactory] = next(
            iter(ep for ep in entrypoints if ep.name == response_factory_name)
        ).load()
        response_factory = response_factory_cls()
        return response_factory

    @classmethod
    def serve(
            cls,
            client_id_key: str, client_secret_key: str,
            consumer_name: str, response_factory_name: str,
            host: str, port: int,
            base_path: str, redirect_path: str,
    ) -> None:

        consumer = cls._get_consumer(consumer_name=consumer_name)
        response_factory = cls._get_response_factory(response_factory_name=response_factory_name)

        app = make_app(
            consumer=consumer,
            response_factory=response_factory,
            notion_client_id=os.environ[client_id_key],
            notion_client_secret=os.environ[client_secret_key],
            base_path=base_path, redirect_path=redirect_path,
        )
        web.run_app(app, host=host, port=port)

    @classmethod
    def mock(cls, host: str, port: int) -> None:
        app = make_mock_app()
        web.run_app(app, host=host, port=port)

    @classmethod
    def _print_http_response(cls, response: web.Response) -> None:
        print(f'Status: {response.status}')
        print(f'Headers:')
        for header_name, header_value in response.headers.items():
            print(f'    {header_name}: {header_value}')
        print(f'Body:\n{response.body.decode()}')

    @classmethod
    def response_auth(
            cls,
            response_factory_name: str,
            access_token: str,
            workspace_id: str,
            workspace_name: str,
            workspace_icon: str,
            bot_id: str,
            owner: str,
    ) -> None:
        response_factory = cls._get_response_factory(response_factory_name=response_factory_name)
        token_info = TokenResponseInfo(
            access_token=access_token, workspace_id=workspace_id,
            workspace_name=workspace_name, workspace_icon=workspace_icon,
            bot_id=bot_id, owner=json.loads(owner),
        )
        response = asyncio.run(response_factory.make_auth_response(token_info=token_info))
        cls._print_http_response(response)

    @classmethod
    def response_error(
            cls,
            response_factory_name: str,
            error: str,
    ) -> None:
        response_factory = cls._get_response_factory(response_factory_name=response_factory_name)
        response = asyncio.run(response_factory.make_error_response(error=error))
        cls._print_http_response(response)

    @classmethod
    def run(cls, args: Any) -> None:
        if args.command == 'serve':
            cls.serve(
                client_id_key=args.client_id_key,
                client_secret_key=args.client_secret_key,
                consumer_name=args.consumer,
                response_factory_name=args.response_factory,
                host=args.host, port=args.port,
                base_path=args.base_path, redirect_path=args.redirect_path,
            )
        elif args.command == 'mock':
            cls.mock(host=args.host, port=args.port)
        elif args.command == 'consumer':
            if args.consumer_command == 'list':
                cls.consumer_list()
        elif args.command == 'response_factory':
            if args.response_factory_command == 'list':
                cls.response_factory_list()
        elif args.command == 'response':
            if args.response_command == 'auth':
                cls.response_auth(
                    response_factory_name=args.response_factory,
                    access_token=args.access_token, workspace_id=args.workspace_id,
                    workspace_name=args.workspace_name, workspace_icon=args.workspace_icon,
                    bot_id=args.bot_id, owner=args.owner,
                )
            elif args.response_command == 'error':
                cls.response_error(
                    response_factory_name=args.response_factory,
                    error=args.error,
                )


def configure_logging() -> None:
    logging.basicConfig(
        level=logging.INFO,
        handlers=[logging.StreamHandler()],
    )


def run() -> None:
    configure_logging()
    NotionOAuthTool.run(args=get_parser().parse_args())



import base64
from http import HTTPStatus
from typing import Any, ClassVar

import aiohttp
import attr
import yarl

import notion_oauth_handler.core.exc as exc
from notion_oauth_handler.core.dto import AuthRedirectInfo, TokenResponseInfo
from notion_oauth_handler.core.consumer import NotionOAuthConsumer


@attr.s
class NotionOAuthHandler:
    """
    Handles Notion OAuth 2
    """

    _default_base_url: ClassVar[str] = 'https://api.notion.com'
    _auth_entrypoint: ClassVar[str] = 'v1/oauth/token'

    _consumer: NotionOAuthConsumer = attr.ib(kw_only=True)
    _client_id: str = attr.ib(kw_only=True)
    _client_secret: str = attr.ib(kw_only=True)
    _base_url: str = attr.ib(kw_only=True)

    @_base_url.default
    def _make_base_url(self) -> str:
        return self._default_base_url

    def _make_token_url(self, redirect_info: AuthRedirectInfo) -> yarl.URL:
        return yarl.URL(self._base_url.rstrip('/')) / self._auth_entrypoint.lstrip('/')

    def _make_token_headers(self, redirect_info: AuthRedirectInfo) -> Any:
        return {
            'Authorization': base64.b64encode(f'{self._client_id}:{self._client_secret}'.encode()).decode(),
        }

    def _make_token_body(self, redirect_info: AuthRedirectInfo) -> dict:
        return {
            'grant_type': 'authorization_code',
            'code': redirect_info.code,
            'redirect_uri': redirect_info.redirect_uri,
        }

    async def _make_token_request(self, redirect_info: AuthRedirectInfo) -> TokenResponseInfo:
        # Make request parameters
        url = self._make_token_url(redirect_info=redirect_info)
        body = self._make_token_body(redirect_info=redirect_info)
        headers = self._make_token_headers(redirect_info=redirect_info)

        # Make the request
        async with aiohttp.ClientSession() as session:
            async with session.post(url, data=body, headers=headers) as response:
                if response.status != HTTPStatus.OK:
                    raise exc.TokenRequestFailed('Access token request to Notion has failed')
                response_body = await response.json()

        # Pack response into DTO
        token_info = TokenResponseInfo(
            access_token=response_body['access_token'],
            workspace_id=response_body['workspace_id'],
            workspace_name=response_body['workspace_name'],
            workspace_icon=response_body['workspace_icon'],
            bot_id=response_body['bot_id'],
            owner=response_body['owner'],
        )
        return token_info

    async def handle_error(self, error: str) -> None:
        await self._consumer.consume_redirect_error(error=error)
        raise exc.NotionAccessDenied('Notion access was denied')

    async def handle_auth(self, redirect_info: AuthRedirectInfo) -> TokenResponseInfo:
        state_info = await self._consumer.consume_redirect_info(redirect_info=redirect_info)
        token_info = await self._make_token_request(redirect_info=redirect_info)
        await self._consumer.consume_token_info(token_info=token_info, state_info=state_info)
        return token_info



import abc
from http import HTTPStatus

from aiohttp.web import Response

from notion_oauth_handler.core.dto import TokenResponseInfo


class NotionOAuthResponseFactory(abc.ABC):
    @abc.abstractmethod
    async def make_error_response(self, error: str) -> Response:
        raise NotImplementedError

    @abc.abstractmethod
    async def make_auth_response(self, token_info: TokenResponseInfo) -> Response:
        raise NotImplementedError


class DummyNotionOAuthResponseFactory(NotionOAuthResponseFactory):
    async def make_error_response(self, error: str) -> Response:
        return Response(text='Error', status=HTTPStatus.FORBIDDEN)

    async def make_auth_response(self, token_info: TokenResponseInfo) -> Response:
        return Response(text='OK', status=HTTPStatus.OK)


class EchoNotionOAuthResponseFactory(NotionOAuthResponseFactory):
    async def make_error_response(self, error: str) -> Response:
        return Response(text=f'Error: {error}', status=HTTPStatus.FORBIDDEN)

    async def make_auth_response(self, token_info: TokenResponseInfo) -> Response:
        return Response(text=f'Token info: {token_info}', status=HTTPStatus.OK)



from aiohttp import web

from notion_oauth_handler.core.consumer import NotionOAuthConsumer
from notion_oauth_handler.server.view import NotionOAuthRedirectView
from notion_oauth_handler.server.middleware import notion_oauth_middleware_factory
from notion_oauth_handler.server.response import NotionOAuthResponseFactory


def make_app(
        consumer: NotionOAuthConsumer,
        response_factory: NotionOAuthResponseFactory,
        notion_client_id: str,
        notion_client_secret: str,
        base_path: str = '',
        redirect_path: str = '/auth_redirect',
) -> web.Application:
    """
    Create Notion OAuth handling server (aiohttp Application)

    :param consumer: a `NotionOAuthConsumer` subclass instance
        that implements the main logic (e.g. stores the token in a database)
    :param response_factory: a `NotionOAuthResponseFactory` subclass instance
        that renders HTTP responses of the server
    :param notion_client_id: client ID of the Notion integration
    :param notion_client_secret: client secret of the Notion integration
    :param base_path: custom base path for all routes
    :param redirect_path: relative path for the redirect handler
    :return: an aiohttp `Application` instance
    """

    base_path = base_path.rstrip('/')
    redirect_path = redirect_path.lstrip('/')

    app = web.Application(
        middlewares=[
            notion_oauth_middleware_factory(
                consumer=consumer,
                response_factory=response_factory,
                notion_client_id=notion_client_id,
                notion_client_secret=notion_client_secret,
            )
        ],
    )
    app.add_routes([
        web.get(f'{base_path}/{redirect_path}', NotionOAuthRedirectView),
    ])
    return app



from typing import Awaitable, Callable

from aiohttp.web import middleware, Request, Response

from notion_oauth_handler.core.consumer import NotionOAuthConsumer
from notion_oauth_handler.core.oauth_handler import NotionOAuthHandler
from notion_oauth_handler.server.response import NotionOAuthResponseFactory


OAUTH_HANDLER_REQUEST_KEY = '__ouath_handler__'
RESPONSE_FACTORY_REQUEST_KEY = '__response_factory__'


def notion_oauth_middleware_factory(
        notion_client_id: str,
        notion_client_secret: str,
        consumer: NotionOAuthConsumer,
        response_factory: NotionOAuthResponseFactory,
):
    oauth_handler = NotionOAuthHandler(
        consumer=consumer,
        client_id=notion_client_id, client_secret=notion_client_secret,
    )

    @middleware
    async def middleware_impl(request: Request, handler: Callable[[Request], Awaitable[Response]]) -> Response:
        request[OAUTH_HANDLER_REQUEST_KEY] = oauth_handler
        request[RESPONSE_FACTORY_REQUEST_KEY] = response_factory
        return await handler(request)

    return middleware_impl



import logging
import uuid

from aiohttp import web


_LOGGER = logging.getLogger(__name__)


class MockTokenView(web.View):
    async def get(self) -> web.Response:
        # TODO: Implement me
        _LOGGER.info('Mock: accepted token request')
        return web.json_response({
            'access_token': str(uuid.uuid4()),
            'workspace_id': str(uuid.uuid4()),
            'workspace_name': str(uuid.uuid4()),
            'workspace_icon': f'http://{uuid.uuid4()}.png',
            'bot_id': str(uuid.uuid4()),
            'owner': {},
        })


def make_mock_app() -> web.Application:
    app = web.Application(
        middlewares=[],
    )
    app.add_routes([
        web.get('/v1/oauth/token', MockTokenView),
    ])
    return app
